"""Socket capable client extension implementation."""
import socket

from paho.mqtt import client as _client


class Client(_client.Client):
    """UNIX socket capable paho mqtt client."""

    # pylint: disable=redefined-outer-name,too-many-instance-attributes

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._socket = None

    def sock_connect(
        self,
        socket,
        keepalive=60,
        clean_start=_client.MQTT_CLEAN_START_FIRST_ONLY,
        properties=None,
    ):
        """Connect to a remote broker via UNIX socket.

        socket is the path to broker unix socket.
        keepalive: Maximum period in seconds between communications with the
        broker. If no other messages are being exchanged, this controls the
        rate at which the client will send ping messages to the broker.
        clean_start: (MQTT v5.0 only) True, False or MQTT_CLEAN_START_FIRST_ONLY.
        Sets the MQTT v5.0 clean_start flag always, never or on the first successful connect only,
        respectively.  MQTT session data (such as outstanding messages and subscriptions)
        is cleared on successful connect when the clean_start flag is set.
        properties: (MQTT v5.0 only) the MQTT v5.0 properties to be sent in the
        MQTT"""

        if self._protocol == _client.MQTTv5:  # pragma: no cover
            self._mqttv5_first_connect = True
        else:  # pragma: no cover
            if clean_start != _client.MQTT_CLEAN_START_FIRST_ONLY:
                raise ValueError("Clean start only applies to MQTT V5")
            if properties is not None:
                raise ValueError("Properties only apply to MQTT V5")

        self.sock_connect_async(socket, keepalive, clean_start, properties)
        return self.reconnect()

    def sock_connect_async(
        self,
        socket,
        keepalive=60,
        clean_start=_client.MQTT_CLEAN_START_FIRST_ONLY,
        properties=None,
    ):
        """Connect to a remote broker asynchronously using UNIX domain socket. This is a
        non-blocking connect call that can be used with loop_start() to provide very quick
        start.
        socket is the path to broker unix socket.
        keepalive: Maximum period in seconds between communications with the
        broker. If no other messages are being exchanged, this controls the
        rate at which the client will send ping messages to the broker.
        clean_start: (MQTT v5.0 only) True, False or MQTT_CLEAN_START_FIRST_ONLY.
        Sets the MQTT v5.0 clean_start flag always, never or on the first successful connect only,
        respectively.  MQTT session data (such as outstanding messages and subscriptions)
        is cleared on successful connect when the clean_start flag is set.
        properties: (MQTT v5.0 only) the MQTT v5.0 properties to be sent in the
        MQTT connect packet.  Use the Properties class.
        """
        if keepalive < 0:  # pragma: no cover
            raise ValueError("Keepalive must be >=0.")
        if self._ssl:
            raise NotImplementedError(
                "TLS support for AF_UNIX sockets is not implemented."
            )
        if self._transport != "tcp":
            raise NotImplementedError(
                "Only tcp transport is supported for UNIX socket connections."
            )

        self._socket = socket
        self._keepalive = keepalive
        self._clean_start = (  # pylint: disable=attribute-defined-outside-init
            clean_start
        )
        self._connect_properties = properties
        self._state = _client.mqtt_cs_connect_async

        # We need to pas some junk data here, since for some unknown reason reconnect checks
        # validity of those field on each reconnect, even though they are already validated by
        # connect_async.
        self._host = "undefined"
        self._port = 9

    def _create_socket_connection(self):
        if self._socket is None:
            return super()._create_socket_connection()

        # Proxy servers are not applicable in AF_UNIX, so we silently ignore them.

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(self._socket)
        except OSError as ex:
            # UNIX socket exceptions are much richer than TCP/IP ones - we will narrow it down to
            # ConnectionError for better compatibility with clients prepared just for TCP/IP.
            raise ConnectionError("Socket connection failed.") from ex
        return sock



from unittest.mock import patch

import pytest

import paho_socket
import paho_socket.client


class TestSockConnectAsync:
    def test_on_tls(self):
        client = paho_socket.Client()
        client.tls_set()

        with pytest.raises(NotImplementedError):
            client.sock_connect_async("socket_path")

    def test_websocket(self):
        client = paho_socket.Client(transport="websockets")

        with pytest.raises(NotImplementedError):
            client.sock_connect_async("socket_path")

    def test_success(self):
        client = paho_socket.Client()
        client.sock_connect_async("socket_path")
        assert client._socket == "socket_path"


@patch("paho_socket.client.socket")
class TestCreateSocketConnection:
    def test_fallback_to_super_on_tcpip(self, _):
        client = paho_socket.Client()
        with patch.object(
            paho_socket.client._client.Client, "_create_socket_connection"
        ) as super_create_socket_connection:
            assert (
                client._create_socket_connection()
                is super_create_socket_connection.return_value
            )

    def test_on_socket(self, socket):
        client = paho_socket.Client()
        client.sock_connect_async("socket_path")
        assert client._create_socket_connection() is socket.socket.return_value
        socket.socket.assert_called_with(socket.AF_UNIX, socket.SOCK_STREAM)
        socket.socket.return_value.connect.assert_called_with("socket_path")

    def test_on_socket_error(self, socket):
        client = paho_socket.Client()
        client.sock_connect_async("socket_path")
        socket.socket.return_value.connect.side_effect = OSError()

        with pytest.raises(ConnectionError):
            client._create_socket_connection()


def test_sock_connect():
    client = paho_socket.Client()
    with patch.object(
        paho_socket.client._client.Client, "reconnect"
    ) as super_reconnect:
        client.sock_connect("socket_path")
    assert client._socket == "socket_path"
    super_reconnect.assert_called()



# pylint: disable=missing-module-docstring,invalid-name
from textwrap import dedent

from setuptools import find_packages, setup

with open("requirements.txt", "r") as f:
    requirements = list(
        filter(lambda line: not line.startswith("--"), f.read().splitlines())
    )

try:
    with open("requirements-dev.txt", "r") as f:
        requirements_dev = list(
            filter(lambda line: not line.startswith("--"), f.read().splitlines())
        )
except FileNotFoundError:
    requirements_dev = []

with open("README.md", "r") as f:
    readme = f.read()

with open("VERSION", "r") as f:
    version = f.read().strip()

with open("src/paho_socket/version.py", "w") as f:
    f.write(
        dedent(
            """\
            # pylint: disable=missing-docstring
            __version__ = "{version}"
            """.format(
                version=version
            )
        )
    )

setup(
    name="paho-socket",
    version=version,
    description=(
        "Thin layer built on top of paho-mqtt allowing for connections with unix socket "
        "brokers."
    ),
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Michał Getka",
    author_email="michal.getka@gmail.com",
    url="https://github.com/mgetka/paho-socket",
    python_requires=">=3.7",
    include_package_data=True,
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=requirements,
    keywords="paho unix",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved",
        "Operating System :: POSIX",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Communications",
        "Topic :: Internet",
    ],
    extras_require={"dev": requirements_dev},
)



from time import sleep

import paho.mqtt.client
import pytest

import paho_socket


@pytest.mark.slow
def test_communication(broker):

    recv_socket_message = False
    recv_tcp_message = False

    def socket_message(mqttc, obj, msg):
        nonlocal recv_tcp_message
        if msg.payload == b"tcp_message":
            recv_tcp_message = True

    def tcp_message(mqttc, obj, msg):
        nonlocal recv_socket_message
        if msg.payload == b"socket_message":
            recv_socket_message = True

    tcp = paho.mqtt.client.Client()
    tcp.on_message = tcp_message
    tcp.connect("127.0.0.1", 8520)

    socket = paho_socket.Client()
    socket.on_message = socket_message
    socket.sock_connect("/tmp/paho_socket_test.sock")

    tcp.subscribe("tcp")
    tcp.loop()

    socket.subscribe("socket")
    socket.loop()

    tcp.publish("socket", b"tcp_message")
    tcp.loop()
    socket.loop()
    socket.publish("tcp", b"socket_message")
    socket.loop()
    tcp.loop()

    assert recv_socket_message
    assert recv_tcp_message
